import functools
from copy import deepcopy

from pypika import JoinType, Criterion, Table

from metrics_layer.core.sql.query_base import MetricsLayerQueryBase
from metrics_layer.core.exceptions import QueryError
from metrics_layer.core.model.filter import LiteralValueCriterion, FilterInterval
from metrics_layer.core.model.field import Field
from metrics_layer.core.sql.query_design import MetricsLayerDesign
from metrics_layer.core.sql.query_dialect import query_lookup
from metrics_layer.core.sql.query_generator import MetricsLayerQuery
from metrics_layer.core.sql.query_filter import MetricsLayerFilter


class FunnelQuery(MetricsLayerQueryBase):
    """ """

    def __init__(self, definition: dict, design: MetricsLayerDesign, suppress_warnings: bool = False) -> None:
        self.design = design
        self.query_type = self.design.query_type
        self.no_group_by = self.design.no_group_by
        self.query_lookup = query_lookup
        self.suppress_warnings = suppress_warnings

        self.step_1_time = "step_1_time"
        self.result_cte_name = "result_cte"
        self.base_cte_name = design.base_cte_name
        super().__init__(definition)

    def get_query(self, semicolon: bool = True):
        base_cte_query = self._base_query()

        query = self.get_funnel_base()
        base_cte_query = base_cte_query.with_(Table(query), self.base_cte_name)

        for i, step in enumerate(self.funnel["steps"]):
            previous_step_number = i
            step_number = i + 1

            from_query = self._base_query()
            base_table = Table(self.base_cte_name)
            from_query = from_query.from_(base_table)

            where = self.where_for_event(step, step_number, self.event_date_alias)
            if previous_step_number == 0:
                from_query = self.get_step_1_cte(from_query, base_table)
            else:
                from_query = self.get_step_n_cte(from_query, base_table, previous_step_number)

            from_query = from_query.where(Criterion.all(where))
            base_cte_query = base_cte_query.with_(Table(from_query), self._cte(step_number))

            select, group_by = self.get_select_and_group_by(step_number)

            union_base = self._base_query()
            if step_number == 1:
                union_cte = union_base.from_(Table(self._cte(step_number))).select(*select).groupby(*group_by)
            else:
                union_cte = union_cte.union_all(
                    union_base.from_(Table(self._cte(step_number))).select(*select).groupby(*group_by)
                )

        base_cte_query = base_cte_query.with_(Table(union_cte), self.result_cte_name)

        result_table = Table(self.result_cte_name)
        base_cte_query = base_cte_query.from_(result_table).select(result_table.star)

        if self.having:
            where = self.get_where_from_having(project=self.design)
            base_cte_query = base_cte_query.where(Criterion.all(where))

        base_cte_query = base_cte_query.limit(self.limit)
        sql = str(base_cte_query)
        if semicolon:
            sql += ";"
        return sql

    def get_select_and_group_by(self, step_number: int):
        step_select = self._get_step_select(step_number)
        base_select, group_by = self._get_base_select()
        return step_select + base_select, group_by

    @functools.lru_cache(maxsize=None)
    def _get_base_select(self):
        select = []
        group_by = []
        pk = self.design.functional_pk()
        for field_name in self.metrics + self.dimensions:
            field = self.design.get_field(field_name)
            field_sql = field.sql_query(query_type=self.query_type, functional_pk=pk, alias_only=True)
            if field.field_type != "measure":
                group_by.append(self.sql(field_sql))
            select.append(self.sql(field_sql, alias=field.alias(with_view=True)))
        return select, group_by

    def _get_step_select(self, step_number: int):
        return [
            self.sql(f"'Step {step_number}'", alias="step"),
            self.sql(f"{step_number}", alias="step_order"),
        ]

    def get_step_1_cte(self, from_query, base_table):
        return from_query.select(base_table.star, self.sql(self.event_date_alias, alias=self.step_1_time))

    def get_step_n_cte(self, from_query, base_table, previous_step_number: int):
        prev_cte = self._cte(previous_step_number)
        match_person = f"{self.base_cte_name}.{self.link_alias}={prev_cte}.{self.link_alias}"
        valid_sequence = f"{prev_cte}.{self.step_1_time}<={self.base_cte_name}.{self.event_date_alias}"
        criteria = LiteralValueCriterion(f"{match_person} and {valid_sequence}")

        from_query = from_query.join(Table(prev_cte), JoinType.inner).on(criteria)
        step_1_time = self.sql(f"{prev_cte}.{self.step_1_time}", alias=self.step_1_time)
        return from_query.select(base_table.star, step_1_time)

    def get_funnel_base(self):
        event_date = self.get_event_date()
        self.event_date_alias = event_date.alias(with_view=True)

        link_field = self.design.project.get_field_by_tag("customer")
        self.link_alias = link_field.alias(with_view=True)

        event_condition_fields = self._get_event_condition_fields()

        having_metrics = list(set(f["field"] for f in self.having)) if self.having else []

        dimensions = self.dimensions + [event_date.id(), link_field.id()] + event_condition_fields
        self.metrics = self.metrics + having_metrics
        query = self._subquery(self.metrics, dimensions, self.where, no_group_by=True)
        return query

    def _get_event_condition_fields(self):
        fields = []
        for step in self.funnel["steps"]:
            if isinstance(step, list):
                for f in step:
                    fields.append(f["field"])
            else:
                fields.extend(self.parse_identifiers_from_clause(step))

        return list(set(fields))

    def get_event_date(self):
        if "view_name" in self.funnel:
            view = self.design.get_view(self.funnel["view_name"])
            return self.design.get_field(f"{view.name}.{view.default_date}_raw")

        dates = []
        for metric_name in self.metrics:
            metric = self.design.get_field(metric_name)
            if "." in metric.view.default_date:
                dates.append(tuple(metric.view.default_date.split(".")))
            else:
                dates.append((metric.view.name, metric.view.default_date))

        if len(list(dates)) == 1:
            date_key = f"{list(dates)[0][0]}.{list(dates)[0][-1]}_raw"
            return self.design.get_field(date_key)
        raise QueryError(f"Could not determine event date for funnel: {self._definition}")

    @staticmethod
    def _cte(step_number: int):
        return f"step_{step_number}"

    def where_for_event(self, step: list, step_number: int, event_date_alias: str):
        where = []
        if isinstance(step, list):
            for condition in step:
                where_field = self.design.get_field(condition["field"])
                where_condition = deepcopy(condition)
                where_condition["query_type"] = self.query_type
                f = MetricsLayerFilter(definition=where_condition, design=None, filter_type="where")
                where_field_sql = where_field.sql_query(query_type=self.query_type, alias_only=True)
                where.append(f.criterion(f"{self.base_cte_name}.{where_field_sql}"))
        else:
            f = MetricsLayerFilter(
                definition={"literal": step, "query_type": self.query_type},
                design=self.design,
                filter_type="where",
            )
            where.append(f.sql_query())

        if step_number > 1:
            where.append(self._within_where(event_date_alias, step_number))
        return where

    def _within_where(self, event_date_alias: str, step_number: int):
        unit = FilterInterval.plural(self.funnel["within"]["unit"])
        value = int(self.funnel["within"]["value"])
        start = f"{self._cte(step_number-1)}.{self.step_1_time}"
        end = f"{self.base_cte_name}.{event_date_alias}"

        date_diff = Field.dimension_group_duration_sql(
            start, end, query_type=self.query_type, dimension_group=unit
        )
        return LiteralValueCriterion(f"{date_diff} <= {value}")

    def _subquery(self, metrics: list, dimensions: list, where: list, no_group_by: bool):
        sub_definition = deepcopy(self._definition)
        dimensions_to_add = self._enrich_query_design(metrics, dimensions)
        sub_definition["metrics"] = sorted(metrics)
        sub_definition["dimensions"] = sorted(dimensions + dimensions_to_add)
        sub_definition["where"] = where
        sub_definition["having"] = []
        sub_definition["limit"] = None

        self.design.no_group_by = no_group_by
        query_generator = MetricsLayerQuery(
            sub_definition, design=self.design, suppress_warnings=self.suppress_warnings
        )
        query = query_generator.get_query(semicolon=False)
        self.design.no_group_by = False

        return query

    def _enrich_query_design(self, metrics: list, dimensions: list):
        dimensions_to_add = []
        for dimension_name in dimensions:
            dimension = self.design.get_field(dimension_name)
            if dimension.id() not in self.design.field_lookup:
                self.design.field_lookup[dimension.id()] = dimension

        for metric_name in metrics:
            metric = self.design.get_field(metric_name)
            if metric.view.primary_key.id() not in self.design.field_lookup:
                self.design.field_lookup[metric.view.primary_key.id()] = metric.view.primary_key
                dimensions_to_add.append(metric.view.primary_key.id())

            if metric.id() not in self.design.field_lookup:
                self.design.field_lookup[metric.id()] = metric

        return dimensions_to_add



from metrics_layer.core.exceptions import QueryError, JoinError
from metrics_layer.core.parse.config import MetricsLayerConfiguration
from metrics_layer.core.sql.single_query_resolve import SingleSQLQueryResolver
from metrics_layer.core.sql.merged_query_resolve import MergedSQLQueryResolver
from metrics_layer.core.sql.query_base import QueryKindTypes


class SQLQueryResolver(SingleSQLQueryResolver):
    """
    Method of resolving the explore name:
        if there is not explore passed (using the format explore_name.field_name), we'll search for
        just the field name and iff that field is used in only one explore, set that as the active explore.
            - Any fields specified that are not in that explore will raise an error

        if it's passed explicitly, use the first metric's explore, and raise an error if anything conflicts
        with that explore
    """

    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        funnel: dict = {},  # A dict with steps (list) and within (string)
        where: str = None,  # Either a list of json or a string
        having: str = None,  # Either a list of json or a string
        order_by: str = None,  # Either a list of json or a string
        config: MetricsLayerConfiguration = None,
        **kwargs,
    ):
        self.field_lookup = {}
        self.no_group_by = False
        self.verbose = kwargs.get("verbose", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.explore_name = kwargs.get("explore_name")
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.config = config
        self.project = self.config.project
        self.metrics = metrics
        self.dimensions = dimensions
        self.funnel = funnel
        self.where = where
        self.having = having
        self.order_by = order_by
        self.kwargs = kwargs
        self.connection = None

        model_name = self.kwargs.get("model_name")
        models = self.project.models()

        # If you specify the model that's top priority
        if model_name:
            self.model = self.project.get_model(model_name)
        # Otherwise, if there's only one option, we use that
        elif len(models) == 1:
            self.model = models[0]
        # Finally, check views for models
        else:
            self.model = self._derive_model()

    @property
    def is_merged_result(self):
        has_explicit_merge = any(self.project.get_field(m).is_merged_result for m in self.metrics)
        has_specified_merge = self.kwargs.get("merged_result", False)
        return has_explicit_merge or has_specified_merge

    def get_query(self, semicolon: bool = True):
        is_explicit_merge = self.is_merged_result
        if not is_explicit_merge:
            try:
                self.query_kind = QueryKindTypes.single
                return self._get_single_query(semicolon=semicolon)
            except JoinError:
                if self.verbose:
                    print("Could not execute the query as a single query. Trying as a merged result query.")
        self.query_kind = QueryKindTypes.merged
        return self._get_merged_result_query(semicolon=semicolon)

    def _get_single_query(self, semicolon: bool):
        resolver = SingleSQLQueryResolver(
            metrics=self.metrics,
            dimensions=self.dimensions,
            funnel=self.funnel,
            where=self.where,
            having=self.having,
            order_by=self.order_by,
            model=self.model,
            config=self.config,
            **self.kwargs,
        )
        query = resolver.get_query(semicolon)
        self.connection = resolver.connection
        return query

    def _get_merged_result_query(self, semicolon: bool):
        resolver = MergedSQLQueryResolver(
            metrics=self.metrics,
            dimensions=self.dimensions,
            funnel=self.funnel,
            where=self.where,
            having=self.having,
            order_by=self.order_by,
            model=self.model,
            config=self.config,
            **self.kwargs,
        )
        query = resolver.get_query(semicolon)
        self.connection = resolver.connection
        return query

    def _derive_model(self):
        all_fields = self.metrics + self.dimensions
        all_model_names = {f.view.model_name for f in all_fields}

        if len(all_model_names) == 0:
            raise QueryError(
                "No models found in this data model. Please specify a model "
                "to connect a data warehouse to your data model."
            )
        elif len(all_model_names) == 1:
            return self.project.get_model(list(all_model_names)[0])
        else:
            raise QueryError(
                "More than one model found in this data model. Please specify a model "
                "to use by either passing the name of the model using 'model_name' parameter or by  "
                "setting the `model_name` property on the view."
            )



import networkx
from itertools import combinations

from metrics_layer.core.exceptions import QueryError
from collections import defaultdict
from copy import deepcopy
from .base import SQLReplacement
from .join import Join


class IdentifierTypes:
    primary = "primary"
    foreign = "foreign"
    join = "join"


class JoinGraph(SQLReplacement):
    def __init__(self, project) -> None:
        self.project = project
        self._join_preference = ["one_to_one", "many_to_one", "one_to_many", "many_to_many"]
        self._merged_result_graph = None
        self._graph = None

    def subgraph(self, view_names: list):
        return self.graph.subgraph(view_names)

    @property
    def graph(self):
        if self._graph is None:
            self._graph = self.build()
        return self._graph

    def join_graph_hash(self, view_name: str) -> str:
        graph = self.project.join_graph.graph
        sorted_components = self._strongly_connected_components(graph)

        for i, components in enumerate(sorted_components):
            if view_name in components:
                return f"subquery_{i}"
        raise QueryError(
            f"View name {view_name} not found in any joinable part of your data model. "
            "Please make sure this is the right name for the view."
        )

    def weak_join_graph_hashes(self, view_name: str) -> list:
        graph = self.project.join_graph.graph
        sorted_components = self._strongly_connected_components(graph)

        join_graph_hashes = []
        for i, components in enumerate(sorted_components):
            subgraph_nodes = self._subgraph_nodes_from_components(graph, components)
            if view_name in subgraph_nodes:
                join_graph_hashes.append(f"subquery_{i}")

        return join_graph_hashes

    def _strongly_connected_components(self, graph):
        components = networkx.strongly_connected_components(graph)
        sorted_components = sorted(components, key=len, reverse=True)  # Sort by largest component
        return sorted_components

    @staticmethod
    def _subgraph_nodes_from_components(graph, components):
        edges = networkx.edge_dfs(graph, source=components)
        all_edges = list(edges) + [list(components)]
        return list(set(node for edge in all_edges for node in edge))

    def ordered_joins(self, view_pairs: list):
        joins = []
        for base_view, join_view in view_pairs:
            join = self.get_join(base_view, join_view)
            joins.append(join)
        return joins

    def collect_errors(self):
        errors = []
        for join in self.joins():
            errors.extend(join.collect_errors())
        return errors

    def joins(self):
        joins = []
        for base_view_name, join_view_name in self.graph.edges():
            joins.append(self.get_join(base_view_name, join_view_name))
        return joins

    def get_join(self, base_view_name: str, join_view_name: str):
        join_info = self.graph[base_view_name][join_view_name]
        join_definition = {**join_info, "base_view_name": base_view_name, "join_view_name": join_view_name}
        return Join(join_definition, project=self.project)

    def build(self):
        graph = networkx.DiGraph()
        identifier_map = self._identifier_map()
        reference_map = self._reference_map()
        for view in self.project.views():
            graph.add_node(view.name)
            if view.name in reference_map:
                # Add all explicit "join" type references
                for join_view_name, join_identifier in reference_map[view.name].items():
                    graph.add_edge(view.name, join_view_name, **join_identifier)

            for identifier in view.identifiers:
                only_join = identifier.get("only_join", [])
                # Add all identifier matches across other views
                for join_view_name in identifier_map.get(identifier["name"], []):

                    if join_view_name != view.name and self._allowed_join(only_join, join_view_name):
                        join_view = self.project.get_view(join_view_name)
                        join_identifier = join_view.get_identifier(identifier["name"])
                        join_only_join = join_identifier.get("only_join", [])
                        if not self._allowed_join(join_only_join, view.name):
                            continue

                        join_info = self._identifier_to_join(
                            first_identifier=identifier,
                            first_view_name=view.name,
                            second_identifier=join_identifier,
                            second_view_name=join_view_name,
                        )
                        is_fanout = join_info["relationship"] in {"one_to_many", "many_to_many"}
                        if is_fanout and join_view_name not in identifier.get("allowed_fanouts", []):
                            continue

                        # Make sure the new join is preferable to the old one
                        if graph.has_edge(view.name, join_view_name):
                            existing = graph[view.name][join_view_name]["relationship"]
                            existing_score = self._join_preference.index(existing)
                            new_score = self._join_preference.index(join_info["relationship"])
                            # Only if the new identifier gives us a more preferable join will we change
                            if existing_score > new_score:
                                graph.add_edge(view.name, join_view_name, **join_info)
                        else:
                            graph.add_edge(view.name, join_view_name, **join_info)

        # print(networkx.to_dict_of_dicts(graph))
        return graph

    def merged_results_graph(self, model):
        if self._merged_result_graph is None:
            with_dates = [
                field
                for field in self.project.fields(model=model)
                if field.canon_date and field.field_type == "measure"
            ]
            mappings = model.get_mappings()

            root_node_name = "canon_date_core"
            graph = networkx.DiGraph()
            graph.add_node(root_node_name)
            # Merged result shared date and field mapping
            join_group_hashes = set()

            for measure in with_dates:
                graph.add_edge(root_node_name, measure.id())
                canon_date = self.project.get_field_by_name(measure.canon_date)
                for timeframe in canon_date.timeframes:
                    canon_date.dimension_group = timeframe
                    graph.add_edge(root_node_name, canon_date.id())
                join_group_hashes.add(self.project.join_graph.join_graph_hash(measure.view.name))

            for from_field, mapping in mappings.items():
                to_field = mapping["field"]
                if (
                    mapping["from_join_hash"] in join_group_hashes
                    and mapping["to_join_hash"] in join_group_hashes
                ):
                    graph.add_edge(root_node_name, self.project.get_field(from_field).id())
                    graph.add_edge(root_node_name, self.project.get_field(to_field).id())

            ordered_hashes = sorted(list(join_group_hashes))
            for join_group_hash_1, join_group_hash_2 in combinations(ordered_hashes, 2):
                join_root = join_group_hash_1 + "_" + join_group_hash_2
                graph.add_node(join_root)

                pair = [join_group_hash_1, join_group_hash_2]
                for measure in with_dates:
                    join_hash = self.project.join_graph.join_graph_hash(measure.view.name)
                    if join_hash in pair:
                        graph.add_edge(join_root, measure.id())
                        canon_date = self.project.get_field_by_name(measure.canon_date)
                        for timeframe in canon_date.timeframes:
                            canon_date.dimension_group = timeframe
                            graph.add_edge(join_root, canon_date.id())

                for view in self.project.views(model=model):
                    join_hashes = self.project.join_graph.weak_join_graph_hashes(view.name)
                    if all(join_hash in join_hashes for join_hash in pair):
                        view_fields = self.project.fields(
                            view_name=view.name, model=model, expand_dimension_groups=True
                        )
                        for field in view_fields:
                            graph.add_edge(join_root, field.id())

                for from_field, mapping in mappings.items():
                    to_field = mapping["field"]
                    if mapping["from_join_hash"] in pair and mapping["to_join_hash"] in pair:
                        graph.add_edge(join_root, self.project.get_field(from_field).id())
                        graph.add_edge(join_root, self.project.get_field(to_field).id())

            self._merged_result_graph = graph
        return self._merged_result_graph

    def merged_results_graph_1(self, model):
        if self._merged_result_graph is None:
            merged_results = [field for field in self.project.fields() if field.is_merged_result]
            mappings = model.get_mappings()

            graph = networkx.DiGraph()
            for merged_result in merged_results:
                root_node = merged_result.id()
                graph.add_node(root_node)

                join_group_hashes = []
                for ref_field in merged_result.referenced_fields(merged_result.sql):
                    canon_date = self.project.get_field_by_name(ref_field.canon_date)
                    graph.add_edge(root_node, ref_field.id())
                    for timeframe in canon_date.timeframes:
                        canon_date.dimension_group = timeframe
                        graph.add_edge(root_node, canon_date.id())
                    join_group_hashes.append(self.project.join_graph.join_graph_hash(ref_field.view.name))

                for from_field, mapping in mappings.items():
                    to_field = mapping["field"]
                    if (
                        mapping["from_join_hash"] in join_group_hashes
                        and mapping["to_join_hash"] in join_group_hashes
                    ):
                        graph.add_edge(root_node, self.project.get_field(from_field).id())
                        graph.add_edge(root_node, self.project.get_field(to_field).id())
            self._merged_result_graph = graph
        return self._merged_result_graph

    def _identifier_map(self):
        result = defaultdict(list)
        for view in self.project.views():
            for identifier in view.identifiers:
                if identifier["type"] != IdentifierTypes.join:
                    result[identifier["name"]].append(view.name)
        return result

    def _reference_map(self):
        result = defaultdict(dict)
        for view in self.project.views():
            for identifier in view.identifiers:
                if identifier["type"] == IdentifierTypes.join:
                    join_identifier = deepcopy(identifier)
                    join_identifier["relationship"] = self._invert_relationship(
                        join_identifier["relationship"]
                    )
                    # We need to invert the join here because this is the inverse
                    # direction of how the join was defined
                    result[view.name][identifier["reference"]] = self._verify_identifier_join(identifier)
                    result[identifier["reference"]][view.name] = self._verify_identifier_join(join_identifier)
        return result

    def _identifier_to_join(self, first_identifier, first_view_name, second_identifier, second_view_name):
        relationship = self._derive_relationship(first_identifier, second_identifier)
        join_type = "left_outer"
        first_clause = self._identifier_join_clause(first_identifier, first_view_name)
        second_clause = self._identifier_join_clause(second_identifier, second_view_name)
        sql_on = f"{first_clause}={second_clause}"
        weight = self._edge_weight(relationship)
        return {"relationship": relationship, "type": join_type, "sql_on": sql_on, "weight": weight}

    def _identifier_join_clause(self, identifier: dict, view_name: str):
        if "sql" in identifier:
            cleaned_sql = deepcopy(identifier["sql"])
            for field_name in self.fields_to_replace(identifier["sql"]):
                to_replace = "${" + field_name + "}"
                if field_name != "TABLE" and "." not in field_name:
                    cleaned_reference = "${" + f"{view_name}.{field_name}" + "}"
                    cleaned_sql = cleaned_sql.replace(to_replace, cleaned_reference)
            clause = cleaned_sql
        else:
            clause = "${" + f"{view_name}.{identifier['name']}" + "}"
        return clause

    def _verify_identifier_join(self, join: dict):
        clean_join = deepcopy(join)
        clean_join["type"] = join.get("type", "left_outer")
        clean_join["relationship"] = join.get("relationship", "many_to_one")
        clean_join["sql_on"] = join["sql_on"]
        clean_join["weight"] = self._edge_weight(clean_join["relationship"])
        return clean_join

    @staticmethod
    def _derive_relationship(identifier, join_identifier):
        base_type = identifier["type"]
        join_type = join_identifier["type"]
        if base_type == IdentifierTypes.foreign and join_type == IdentifierTypes.primary:
            return "many_to_one"
        elif base_type == IdentifierTypes.primary and join_type == IdentifierTypes.primary:
            return "one_to_one"
        elif base_type == IdentifierTypes.primary and join_type == IdentifierTypes.foreign:
            return "one_to_many"
        elif base_type == IdentifierTypes.foreign and join_type == IdentifierTypes.foreign:
            return "many_to_many"
        else:
            raise QueryError(
                "This join type cannot be determined from the identifier properties. "
                f"Make sure you've set the properties correctly. Base type: {base_type},"
                f" join type: {join_type}"
            )

    @staticmethod
    def _invert_relationship(relationship: str):
        mapping = {
            "many_to_one": "one_to_many",
            "one_to_many": "many_to_one",
            "many_to_many": "many_to_many",
            "one_to_one": "one_to_one",
        }
        return mapping[relationship]

    @staticmethod
    def _edge_weight(relationship: str):
        mapping = {
            "many_to_one": 2,
            "one_to_many": 3,
            "many_to_many": 4,
            "one_to_one": 1,
        }
        return mapping[relationship]

    @staticmethod
    def _allowed_join(only_join: list, view_name: str):
        return not only_join or (only_join and view_name in only_join)



import os
from copy import deepcopy

from metrics_layer.core.model.project import Project
from metrics_layer.core.parse.connections import (
    BigQueryConnection,
    ConnectionType,
    RedshiftConnection,
    SnowflakeConnection,
)

from .github_repo import GithubRepo, LocalRepo, LookerGithubRepo
from .manifest import Manifest
from .project_reader import ProjectReader


class ConfigError(Exception):
    pass


class MetricsLayerConfiguration:
    def __init__(
        self,
        repo_config: dict = None,
        additional_repo_config: dict = None,
        connections: list = [],
        target: str = None,
    ):
        self.profiles_path = None
        self.env_name = target
        self.looker_env = self.env_name
        self.repo, conns = self._resolve_config(repo_config, prefix="METRICS_LAYER", raise_exception=True)
        self.additional_repo, addtl_conns = self._resolve_config(
            additional_repo_config, prefix="METRICS_LAYER_ADDITIONAL"
        )
        self._connections = self._parse_connections(connections) + conns + addtl_conns
        self._project = None
        self._user = None

    def set_user(self, user: dict):
        self._user = user
        if self._project is not None:
            self._project.set_user(self._user)

    def get_branch_options(self):
        # We can't have branch options with two independent repos
        if self.additional_repo is not None:
            return []
        return self.repo.branch_options

    @staticmethod
    def get_metrics_layer_configuration(config=None, target: str = None):
        if config:
            return config
        return MetricsLayerConfiguration(target=target)

    @property
    def project(self):
        if self._project is None:
            self.load()
        return self._project

    def connections(self):
        return self._connections

    def get_connection(self, connection_name: str):
        try:
            return next((c for c in self.connections() if c.name == connection_name))
        except StopIteration:
            raise ConfigError(f"Could not find connection named {connection_name} in {self.connections()}")

    def add_connection(self, connection_dict: dict):
        parsed_connections = self._parse_connections([connection_dict])
        self._connections.extend(parsed_connections)

    def load(self):
        self._project = self._get_project()
        self._project.set_user(self._user)

    def dump(self, path: str = "./"):
        reader = ProjectReader(repo=self.repo, additional_repo=self.additional_repo)
        reader.load()
        reader.dump(path)

    def _get_project(self):
        if self.profiles_path:
            profiles_dir = os.path.dirname(self.profiles_path)
        else:
            profiles_dir = None
        reader = ProjectReader(
            repo=self.repo, additional_repo=self.additional_repo, profiles_dir=profiles_dir
        )
        reader.load()
        connection_lookup = {c.name: c.type for c in self.connections()}
        project = Project(
            models=reader.models,
            views=reader.views,
            dashboards=reader.dashboards,
            looker_env=self.looker_env,
            connection_lookup=connection_lookup,
            manifest=Manifest(reader.manifest),
        )
        return project

    def _resolve_config(self, config: dict, prefix: str, raise_exception: bool = False):
        # Config is passed explicitly: this gets first priority
        if config is not None and isinstance(config, dict):
            return self._get_config_repo(config), []

        # Next look for environment variables
        repo = self._get_repo_from_environment(prefix)
        if repo:
            return repo, []

        # Finally look for config file
        repo, connections = self._get_repo_from_config_file(prefix, config, target_name=self.env_name)
        if repo:
            return repo, connections

        if raise_exception:
            raise ConfigError(
                """ We could not find a valid configuration in the environment. Try following the
                documentation (TODO) to properly set your environment variables, or pass the
                configuration explicitly as a dictionary
            """
            )
        return None, []

    @staticmethod
    def _get_config_repo(config: dict):
        if all(a in config for a in ["repo_path"]):
            return LocalRepo(**config)

        if all(a in config for a in ["repo_url", "branch"]):
            return GithubRepo(**config)

        if all(a in config for a in ["looker_url", "client_id", "client_secret", "project_name"]):
            return LookerGithubRepo(**config)

        raise ConfigError(
            f"""Could not initialize any known config with the arguments passed:
            {list(config.keys())} We either initialize a local repo with the following arguments:
            'repo_path': 'path/to/my/repo'
            or a direct Github connection with following arguments:
            'repo_url': https://username:personal_access_token@github.com/yourorg/yourrepo.git, 'branch': dev
            or through the Looker API with the following arguments:
            'looker_url': https://company.cloud.looker.com, 'client_id': aefafasdasd,
            'client_secret': adfawdada, 'project_name': yourcompany

            Note: dbt repos require an additional 'warehouse_type' argument
            """
        )

    @staticmethod
    def _get_repo_from_environment(prefix: str):
        warehouse_type = os.getenv(f"{prefix}_WAREHOUSE_TYPE")
        local_repo_param = os.getenv(f"{prefix}_REPO_PATH")
        if local_repo_param:
            repo_type = os.getenv(f"{prefix}_REPO_TYPE")
            return LocalRepo(repo_path=local_repo_param, repo_type=repo_type, warehouse_type=warehouse_type)

        repo_params = [os.getenv(f"{prefix}_{a}") for a in ["REPO_URL", "BRANCH"]]
        if all(repo_params):
            repo_type = os.getenv(f"{prefix}_REPO_TYPE")
            return GithubRepo(
                repo_url=repo_params[0],
                branch=repo_params[1],
                repo_type=repo_type,
                warehouse_type=warehouse_type,
            )

        looker_repo_params = [
            os.getenv(f"{prefix}_{a}") for a in ["LOOKER_URL", "CLIENT_ID", "CLIENT_SECRET", "PROJECT_NAME"]
        ]
        if all(looker_repo_params):
            config = {
                "looker_url": looker_repo_params[0],
                "client_id": looker_repo_params[1],
                "client_secret": looker_repo_params[2],
                "project_name": looker_repo_params[3],
                "repo_type": "lookml",
            }
            return LookerGithubRepo(**config)

    def _get_repo_from_config_file(self, prefix: str, config_profile_name: str, target_name: str = None):
        if config_profile_name is None:
            return None, []

        clean_prefix = "additional_" if "additional" in prefix.lower() else ""

        metrics_layer_profiles_directory = self.get_metrics_layer_profiles_directory()
        self.profiles_path = os.path.join(metrics_layer_profiles_directory, "profiles.yml")

        if not os.path.exists(self.profiles_path):
            raise ConfigError(
                f"""MetricsLayer could not find the profiles.yml file in the directory
             {self.profiles_path}, please check that path to ensure the file exists.
             If this does not look like the path you specified, make sure you haven't set
             the environment variable METRICS_LAYER_PROFILES_DIR to another value
             which may be overriding the default (~, your home directory)"""
            )

        all_profiles = ProjectReader.read_yaml_file(self.profiles_path)

        try:
            profile = all_profiles[config_profile_name]
        except KeyError:
            raise ConfigError(
                f"MetricsLayer could not find the profile "
                f"{config_profile_name} in options {list(all_profiles.keys())} "
                f"located in the profiles.yml file at {self.profiles_path}"
            )

        if "target" not in profile:
            raise ConfigError(
                "You need to specify a default target in the "
                f"file {self.profiles_path} for profile {config_profile_name}"
            )

        if "outputs" not in profile:
            raise ConfigError(
                "You need to specify outputs in your profile so MetricsLayer can "
                f"connect to your database for profile {config_profile_name}"
            )

        try:
            target_name = target_name if target_name else profile["target"]
            target = profile["outputs"][target_name]
        except KeyError:
            raise ConfigError(
                f"MetricsLayer could not find the target "
                f"{target_name} in outputs {list(profile['outputs'].keys())} "
                f"located in the profiles.yml file at {self.profiles_path}"
            )

        if "looker_env" in target:
            self.looker_env = target["looker_env"]

        repo_type = target.get(f"{clean_prefix}repo_type")
        raw_connection = deepcopy(target)
        warehouse_type = raw_connection.get("type")

        # Github repo
        if all(k in target for k in [f"{clean_prefix}repo_url", f"{clean_prefix}branch"]):
            repo = GithubRepo(
                repo_url=target[f"{clean_prefix}repo_url"],
                branch=target[f"{clean_prefix}branch"],
                repo_type=repo_type,
                warehouse_type=warehouse_type,
            )

        # Looker API
        looker_keys = [
            f"{clean_prefix}looker_url",
            f"{clean_prefix}client_id",
            f"{clean_prefix}client_secret",
            f"{clean_prefix}project_name",
        ]
        if all(k in target for k in looker_keys):
            looker_args = {k: target[k] for k in looker_keys}
            repo = LookerGithubRepo(**looker_args, repo_type="lookml")

        # Local repo
        path_arg = f"{clean_prefix}repo_path"
        if path_arg in target:
            if os.path.isabs(target[path_arg]):
                path = target[path_arg]
            else:
                path = os.path.abspath(
                    os.path.join(metrics_layer_profiles_directory, os.path.expanduser(target[path_arg]))
                )
        else:
            config_path = os.path.join(os.getcwd(), "zenlytic_project.yml")
            if os.path.exists(config_path):
                config = ProjectReader.read_yaml_file(config_path)
                relative_path = config.get("folder", "./")
                path = os.path.abspath(relative_path)
                dbt_path = os.getcwd()
            else:
                path = os.getcwd()
                dbt_path = os.getcwd()
        repo = LocalRepo(
            repo_path=path, repo_type=repo_type, dbt_path=dbt_path, warehouse_type=warehouse_type
        )

        connection = {
            **raw_connection,
            "name": config_profile_name,
            "directory": metrics_layer_profiles_directory,
        }
        return repo, MetricsLayerConfiguration._parse_connections([connection])

    @staticmethod
    def get_all_profiles(directory, names_only: bool = False):
        all_profiles = ProjectReader.read_yaml_file(directory)
        if names_only:
            return list(all_profiles.keys())
        return all_profiles

    @staticmethod
    def get_metrics_layer_profiles_directory():
        env_specified_location = os.getenv(f"METRICS_LAYER_PROFILES_DIR")
        if env_specified_location:
            if os.path.isabs(env_specified_location):
                return env_specified_location
            else:
                return os.path.join(os.getcwd(), os.path.abspath(env_specified_location))

        # System default home directory to dbt profiles dir
        home = os.path.expanduser("~")
        location = os.path.join(home, ".dbt/")
        return location

    @staticmethod
    def _parse_connections(connections: list):
        class_lookup = {
            ConnectionType.snowflake: SnowflakeConnection,
            ConnectionType.bigquery: BigQueryConnection,
            ConnectionType.redshift: RedshiftConnection,
        }
        results = []
        for connection in connections:
            connection_type = connection["type"].upper()
            connection_class = class_lookup[connection_type](**connection)
            results.append(connection_class)
        return results



from metrics_layer.core.exceptions import QueryError
from metrics_layer.core.parse.config import MetricsLayerConfiguration
from collections import defaultdict
from copy import deepcopy

from metrics_layer.core.sql.query_merged_results import MetricsLayerMergedResultsQuery
from metrics_layer.core.sql.single_query_resolve import SingleSQLQueryResolver


class MergedSQLQueryResolver(SingleSQLQueryResolver):
    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        funnel: dict = {},
        where: str = None,  # Either a list of json or a string
        having: str = None,  # Either a list of json or a string
        order_by: str = None,  # Either a list of json or a string
        model=None,
        config: MetricsLayerConfiguration = None,
        **kwargs,
    ):
        if funnel != {}:
            raise QueryError("Funnel queries are not supported in merged results queries")

        self.field_lookup = {}
        self.no_group_by = False
        self.has_cumulative_metric = False
        self.verbose = kwargs.get("verbose", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.return_pypika_query = kwargs.get("return_pypika_query")
        self.force_group_by = kwargs.get("force_group_by", False)
        self.kwargs = kwargs
        self.config = config
        self.project = self.config.project
        self.metrics = metrics
        self.dimensions = dimensions
        self.parse_field_names(where, having, order_by)
        self.model = model

    def get_query(self, semicolon: bool = True):
        self.parse_field_names(self.where, self.having, self.order_by)
        self.derive_sub_queries()

        queries_to_join = {}
        for join_hash in self.query_metrics.keys():
            metrics = [f.id() for f in self.query_metrics[join_hash]]
            dimensions = [f.id() for f in self.query_dimensions.get(join_hash, [])]

            # Overwrite the limit arg because these are subqueries
            kws = {**self.kwargs, "limit": None, "return_pypika_query": True}
            resolver = SingleSQLQueryResolver(
                metrics=metrics,
                dimensions=dimensions,
                where=self.query_where[join_hash],
                having=[],
                order_by=[],
                model=self.model,
                config=self.config,
                **kws,
            )
            query = resolver.get_query(semicolon=False)
            queries_to_join[join_hash] = query

        query_config = {
            "merged_metrics": self.merged_metrics,
            "query_metrics": self.query_metrics,
            "query_dimensions": self.query_dimensions,
            "having": self.having,
            "queries_to_join": queries_to_join,
            "join_hashes": list(sorted(self.query_metrics.keys())),
            "query_type": resolver.query_type,
            "limit": self.limit,
            "project": self.project,
        }
        merged_result_query = MetricsLayerMergedResultsQuery(query_config)
        query = merged_result_query.get_query(semicolon=semicolon)

        self.connection = resolver.connection
        return query

    def derive_sub_queries(self):
        self.query_metrics = defaultdict(list)
        self.merged_metrics = []
        self.secondary_metrics = []

        for metric in self.metrics:
            field = self.project.get_field(metric)
            if field.is_merged_result:
                self.merged_metrics.append(field)
            else:
                self.secondary_metrics.append(field)

        for merged_metric in self.merged_metrics:
            for ref_field in merged_metric.referenced_fields(merged_metric.sql):
                if isinstance(ref_field, str):
                    raise QueryError(f"Unable to find the field {ref_field} in the project")

                join_group_hash = self.project.join_graph.join_graph_hash(ref_field.view.name)
                canon_date = ref_field.canon_date.replace(".", "_")
                key = f"{canon_date}__{join_group_hash}"
                self.query_metrics[key].append(ref_field)

        for field in self.secondary_metrics:
            join_group_hash = self.project.join_graph.join_graph_hash(field.view.name)
            canon_date = field.canon_date.replace(".", "_")
            key = f"{canon_date}__{join_group_hash}"
            if key in self.query_metrics:
                already_in_query = any(field.id() in f.id() for f in self.query_metrics[key])
                if not already_in_query:
                    self.query_metrics[key].append(field)
            else:
                self.query_metrics[key].append(field)

        canon_dates = []
        dimension_mapping = defaultdict(list)
        for join_hash, field_set in self.query_metrics.items():
            if len({f.canon_date for f in field_set}) > 1:
                raise NotImplementedError(
                    "Zenlytic does not currently support different canon_date "
                    "values for metrics in the same subquery"
                )
            canon_date = field_set[0].canon_date
            canon_dates.append(canon_date)
            for other_explore_name, other_field_set in self.query_metrics.items():
                if other_explore_name != join_hash:
                    other_canon_date = other_field_set[0].canon_date
                    canon_date_data = {"field": other_canon_date, "from_join_hash": other_explore_name}
                    dimension_mapping[canon_date].append(canon_date_data)

        mappings = self.model.get_mappings()
        for key, map_to in mappings.items():
            for other_join_hash, other_field_set in self.query_metrics.items():
                if map_to["to_join_hash"] in other_join_hash:
                    map_to["from_join_hash"] = other_join_hash
                dimension_mapping[key].append(deepcopy(map_to))

        self.query_dimensions = defaultdict(list)
        for dimension in self.dimensions:
            field = self.project.get_field(dimension)
            join_group_hash = self.project.join_graph.join_graph_hash(field.view.name)
            field_key = f"{field.view.name}.{field.name}"
            if field_key in canon_dates:
                join_hash = f'{field_key.replace(".", "_")}__{join_group_hash}'
                self.query_dimensions[join_hash].append(field)

                dimension_group = field.dimension_group
                for mapping_info in dimension_mapping[field_key]:
                    if mapping_info["from_join_hash"] in self.query_metrics:
                        key = f"{mapping_info['field']}_{dimension_group}"
                        ref_field = self.project.get_field(key)
                        self.query_dimensions[mapping_info["from_join_hash"]].append(ref_field)
            else:
                not_in_metrics = True
                for join_hash in self.query_metrics.keys():
                    if join_group_hash in join_hash:
                        self.query_dimensions[join_hash].append(field)
                        not_in_metrics = False
                    else:
                        if field_key not in dimension_mapping:
                            raise QueryError(
                                f"Could not find mapping from field {field_key} to other views. "
                                "Please add a mapping to your view definition to allow this."
                            )
                        for mapping_info in dimension_mapping[field_key]:
                            if mapping_info["from_join_hash"] in self.query_metrics:

                                ref_field = self.project.get_field(mapping_info["field"])
                                self.query_dimensions[mapping_info["from_join_hash"]].append(ref_field)
                if not_in_metrics:
                    canon_date = field.canon_date.replace(".", "_")
                    key = f"{canon_date}__{join_group_hash}"
                    self.query_metrics[key] = []
                    self.query_dimensions[key].append(field)

        # Get rid of duplicates while keeping order to make joining work properly
        self.query_dimensions = {
            k: sorted(list(set(v)), key=lambda x: v.index(x)) for k, v in self.query_dimensions.items()
        }

        self.query_where = defaultdict(list)
        for where in self.where:
            field = self.project.get_field(where["field"])
            dimension_group = field.dimension_group
            join_group_hash = self.project.join_graph.join_graph_hash(field.view.name)
            for join_hash in self.query_metrics.keys():
                if join_group_hash in join_hash:
                    self.query_where[join_hash].append(where)
                else:
                    key = f"{field.view.name}.{field.name}"
                    for mapping_info in dimension_mapping[key]:
                        if dimension_group:
                            key = f"{mapping_info['field']}_{dimension_group}"
                        else:
                            key = mapping_info["field"]
                        ref_field = self.project.get_field(key)
                        mapped_where = deepcopy(where)
                        mapped_where["field"] = ref_field.id()
                        self.query_where[join_hash].append(mapped_where)
